function [fixout, floatout, rerr, snr, ovrerr, guardout] = ...
    filterboth(Hd,fixin,floatin, testovr);
% function [fixout, floatout, rerr, snr, ovrerr, guardout] = ...
%   filterboth(Hd,fixin,floatin, testovr);
% function to measure roundoff error of a fixed point implementation
% Hd is the dfilt object (fixed point)
% fixin is the fixed point filter input (best if it's quantized before, but 
% doesn't have to be)
% floatin is the floating point filter input.  floatin is for use when you
% are running parallel streams for fixed and floating point implementations
% and Hd is not the first filter.  floatin is an optional input, if it is
% not specified, fixin is used for both filter inputs
% testovr is an optional input flag, true (1) tests for overflow, false (0)
% or not present skips the overflow test (runs lots faster, too)
%
% fixout is the fixed point filter output
% floatout is the floating point filter output
% rerr is the difference between the fixed point implementation and the
% floating point implementation when BOTH are using QUANTIZED coefficients.
% snr is the signal power (based on the floating point signal output) over
% the noise power (based on rerr at output) in dB.  Note that this is
% cumulative for cascaded filters.
% 
% Hopefully it's obvious that if the coefficients are the same the only
% error that should appear is roundoff error (assuming no overflow).
%
% ovrerr is the difference between the fixed point implementation and the
% same filter object with added guardbits (same fraction bits, but larger
% wordlength).
% guardout is the extra guardbit filter output
%
% Hopefully it's also obvious that if the fixed point filter overflows, the
% one with extra guardbits won't and the error will show up in the two 
% outputs.
% Note this only really detects overflow or not, since errors propagate and
% all later outputs will be different even if there's only one overflow.
% Also note that if an internal addition operation overflows in a wrapping
% mode and recovers, the output will not be changed, so no overflow will be
% detected.


% Copyright 2008
% Wayne T. Padgett
% Last Modified - 
% wtp 9/30/04 ece497 started life as roundofferr()
% wtp 10/10/04 modified to allow optional separate floating point input
% wtp 10/18/04 converted to filterboth() to improve efficiency
% wtp 10/6/05 changed rerr and ovrerr to return double instead of fixed pt
% types
% wtp 10/25/05 added support for dffir filter types
% wtp 1/30/08 added support for df1sos

if nargin < 3,
    floatin = fixin;
end
if nargin < 4,
    testovr = 0;
end


if strcmp(Hd.FilterStructure, 'Direct-Form II, Second-Order Sections')
    sosm = Hd.sosmatrix;
    sc = Hd.scalevalues;

    Hdfloat = copy(Hd); % Hdfloat = Hd only makes a new handle to the SAME 
                        % object!
    Hdfloat.arithmetic = 'double';  % coefficients revert to ideal when you
                                    % do this
    Hdfloat.sosmatrix = sosm; % replace ideal values with quantized values,
                              % computed in floating point
    Hdfloat.scalevalues = sc;

elseif strcmp(Hd.FilterStructure, 'Direct-Form I, Second-Order Sections')
    sosm = Hd.sosmatrix;
    sc = Hd.scalevalues;

    Hdfloat = copy(Hd); % Hdfloat = Hd only makes a new handle to the SAME 
                        % object!
    Hdfloat.arithmetic = 'double';  % coefficients revert to ideal when you
                                    % do this
    Hdfloat.sosmatrix = sosm; % replace ideal values with quantized values,
                              % computed in floating point
    Hdfloat.scalevalues = sc;

elseif strcmp(Hd.filterstructure,'Direct-Form FIR')
    bcoef = Hd.Numerator;
    Hdfloat = copy(Hd);
    Hdfloat.arithmetic = 'double';
    Hdfloat.Numerator = bcoef; % put quantized coefficients back into 
                               % floating point filter
    
end; % if DF2SOS else FIR

fixout = filter(Hd,fixin);
floatout = filter(Hdfloat,floatin);
rerr = double(fixout) - floatout;
snr = 10*log10(sum(floatout.^2)/sum(double(rerr).^2));

if testovr,

    Hdguard = copy(Hd); % Hguard = Hd only makes a new handle to the SAME 
                        % object!
    specifyall(Hdguard); % due to a bug in Filter Design Toolbox 3.0, some 
                         % properties can't be written 
                         %(ProductWordLength below) without re-doing the
                         % specifyall()

    %% add extra guardbits to all the internal data paths
    if strcmp(Hd.filterstructure,'Direct-Form II, Second-Order Sections')

        guard = 8;

        Hdguard.InputWordLength  =   Hdguard.InputWordLength + guard;

        Hdguard.StageInputWordLength  =  Hdguard.StageInputWordLength ...
                                       + guard;

        Hdguard.StageOutputWordLength = Hdguard.StageOutputWordLength ...
                                       + guard;

        Hdguard.OutputWordLength =   Hdguard.OutputWordLength + guard;

        Hdguard.StateWordLength =   Hdguard.StateWordLength + guard;

        % copy object has a bug in FDTB 3.0 that won't let this be written 
        % unless you re-do the specifyall above
        Hdguard.ProductWordLength  = Hdguard.ProductWordLength + guard;

        Hdguard.AccumWordLength  =  Hdguard.AccumWordLength + guard;
        
        
    elseif strcmp(Hd.filterstructure,'Direct-Form I, Second-Order Sections')

        guard = 8;

        Hdguard.InputWordLength  =   Hdguard.InputWordLength + guard;

        Hdguard.OutputWordLength =   Hdguard.OutputWordLength + guard;

        Hdguard.NumStateWordLength =   Hdguard.NumStateWordLength + guard;
        Hdguard.DenStateWordLength =   Hdguard.DenStateWordLength + guard;

        % copy object has a bug in FDTB 3.0 that won't let this be written
        % unless you re-do the specifyall above
        Hdguard.ProductWordLength  = Hdguard.ProductWordLength + guard;

        Hdguard.AccumWordLength  =  Hdguard.AccumWordLength + guard;


    elseif strcmp(Hd.filterstructure,'Direct-Form FIR')
        
        guard = 8;

        Hdguard.InputWordLength  =   Hdguard.InputWordLength + guard;

        Hdguard.OutputWordLength =   Hdguard.OutputWordLength + guard;

        % copy object has a bug in FDTB 3.0 that won't let this be written 
        % unless you re-do the specifyall above
        Hdguard.ProductWordLength  = Hdguard.ProductWordLength + guard;

        Hdguard.AccumWordLength  =  Hdguard.AccumWordLength + guard;
    else
        error('I didn''t write this for anything but df1sos, df2sos and dffir (yet)');
    end
    
    
    
    %fixout = filter(Hd,x);
    guardout = filter(Hdguard,fixin);
    ovrerr = double(guardout) - double(fixout);

    if max(abs(double(ovrerr)))>0,
        warning('overflow has occurred')
    end
end % if testovr